/* SF64-PCR radio card (alias PCI RadioLINK) driver for Linux
 *
 * version 0.1.6, 14.7.2004
 *
 * (c) 2002 Martin "Marble" Beranek <marble@matfyz.cz>
 *
 ***************************************************************************
 *
 *     This program is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU General Public License as
 *     published by the Free Software Foundation; either version 2 of
 *     the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 * *************************************************************************
 *
 * some ideas & code from following sources
 *
 *  - SFR64-PCR FreeBSD driver by Vladimir Popov <jumbo@narod.ru>
 *  - SF-16FMR driver by David Sauer, davids@penguin.cz
 *  - original unusable driver :)
 *
 * TODO:
 *
 *  - to read 'tuned' info and real frequency
 *  - try to fix "self tuned radio" feature
 *
 * Notes:
 *
 *  - you have to enable 'Video for Linux'
 *
 */

#include <linux/module.h>	/* Modules 			*/
#include <linux/init.h>		/* Initdata			*/
#include <linux/ioport.h>	/* check_region, request_region	*/
#include <linux/delay.h>	/* udelay			*/
#include <asm/io.h>		/* outb, outb_p			*/
#include <asm/uaccess.h>	/* copy to/from user		*/
#include <linux/videodev.h>	/* kernel radio structs		*/
#include <asm/semaphore.h>
#include <linux/pci.h>
#include <asm/system.h>		/* cli(), *_flags */


// here should be unique ID for this card...
#define VID_HARDWARE_SF64PCR	12345

#define SF64PCR_VENDOR		0x1319
#define SF64PCR_DEVICE		0x0801

struct pcr_device {
	long port;
	unsigned long freq;
	unsigned long lock, stereo, band, search;
        __u32 flags;
};

static int radio_nr = -1;
static int users = 0;
static struct semaphore lock;


/* referencnmi frekvence na karte (v kHz) */
#define SF64PCR_BASEFREQ	10700

/* from 1/16 kHz to the value for TEA5757 */
#define SF64PCR_ENCODE(x)	((((x)/16 + SF64PCR_BASEFREQ)*10 + 62)/125)

/* in 1/16 kHz */
#define SF64PCR_MINFREQ (80*16000)	/*  80MHz */
#define SF64PCR_MAXFREQ (110*16000)	/* 110MHz */

/****** TEA5757 SHIFT REGISTER ***********************************/

#define TEA5757_SEARCH_START		(1<<24) /* 0x1000000 */
#define TEA5757_SEARCH_END		(0<<24) /* 0x0000000 */

/* pcr_device->search */
#define TEA5757_SEARCH_UP		(1<<23) /* 0x0800000 */
#define TEA5757_SEARCH_DOWN		(0<<23) /* 0x0000000 */

/* pcr_device->stereo */
#define TEA5757_STEREO			(0<<22) /* 0x0000000 */
#define TEA5757_MONO			(1<<22) /* 0x0400000 */

/* pcr_device->band */
/* But no oscilator and other parts for AM receiving... :-( */
#define TEA5757_BAND_FM			(0<<20)
#define TEA5757_BAND_MW			(1<<20)
#define TEA5757_BAND_LW			(2<<20)
#define TEA5757_BAND_SW			(3<<20)

/* pcr_device->lock */
#define TEA5757_S005			(0<<16) /* 0x0000000 */ /* > 5 uV */
#define TEA5757_S010			(2<<16) /* 0x0010000 */ /* > 10 uV */
#define TEA5757_S030			(1<<16) /* 0x0020000 */ /* > 30 uV */
#define TEA5757_S150			(3<<16) /* 0x0030000 */ /* > 150 uV */

#define TEA5757_FREQ			0x0007FFF /* mask for freq. setting */

#define PCR_MUTE(dev) ((dev)->flags |= VIDEO_AUDIO_MUTE)
#define PCR_UNMUTE(dev) ((dev)->flags &= ~VIDEO_AUDIO_MUTE)
#define PCR_IS_MUTE(dev) ((dev)->flags & VIDEO_AUDIO_MUTE)



static void pcr_write(struct pcr_device* dev)
{
	int d, i;
	unsigned long freq = SF64PCR_ENCODE(dev->freq), reg = 0ul;


	freq &= TEA5757_FREQ;
	reg = dev->search | dev->band | dev->lock | dev->stereo | freq;

        down(&lock);
	cli();
	outw(0xf800, dev->port);
	udelay(4);

	for(i=24; i>=0; i--) {
		if (reg & (1<<i))
			d = 0x4;
		else
			d = 0x0;
		outw(0xf800 + d, dev->port);
		udelay(4);
		outw(0xf801 + d, dev->port);
		udelay(4);
		outw(0xf800 + d, dev->port);
		udelay(4);
	}
        if PCR_IS_MUTE(dev)
		outw(0xf800, dev->port);
	else
		outw(0xf802, dev->port);
	udelay(4);
	sti();
        up(&lock);

	return;
}

static int pcr_getsigstr(struct pcr_device *dev)
{
	/* TODO */

	return 0xffff;
}

static int pcr_ioctl(struct video_device *dev, unsigned int cmd, void *arg)
{
	struct pcr_device *pcr = dev->priv;
	
	switch(cmd) {
	case VIDIOCGCAP: {
		struct video_capability v;

		strcpy(v.name, "SF64-PCR radio");
		v.type = VID_TYPE_TUNER;
		v.channels = 1;
		v.audios = 0; /* ... I hope that it's OK... */
		v.maxwidth=0;
		v.maxheight=0;
		v.minwidth=0;
		v.minheight=0;
		if (copy_to_user(arg,&v,sizeof(v)))
			return -EFAULT;
		return 0;
	}
	case VIDIOCGTUNER:
	{
		struct video_tuner v;

		if (copy_from_user(&v, arg,sizeof(v))!=0)
			return -EFAULT;
		if (v.tuner)
			return -EINVAL;
		strcpy(v.name, "FM");
		v.rangelow = SF64PCR_MINFREQ;
		v.rangehigh = SF64PCR_MAXFREQ;
		v.flags = pcr->flags &
			(VIDEO_TUNER_LOW | VIDEO_TUNER_STEREO_ON);
		v.mode = VIDEO_MODE_AUTO;
		v.signal = pcr_getsigstr(pcr);
		if (copy_to_user(arg, &v, sizeof(v)))
			return -EFAULT;
		return 0;
	}
	case VIDIOCSTUNER:
	{
		struct video_tuner v;

		if (copy_from_user(&v, arg, sizeof(v)))
			return -EFAULT;
		if (v.tuner!=0)
			return -EINVAL;
		return 0;
	}
	case VIDIOCGFREQ:
	{
		if (copy_to_user(arg, &(pcr->freq), sizeof(pcr->freq)))
			return -EFAULT;
		return 0;
	}
	case VIDIOCSFREQ:
	{
		if (copy_from_user(&(pcr->freq), arg, sizeof(pcr->freq)))
			return -EFAULT;
		pcr_write(pcr);
		return 0;
	}
	case VIDIOCGAUDIO:
	{	
		struct video_audio v;
		
		v.audio = 0;
		if (pcr->flags & VIDEO_AUDIO_MUTE)
			v.volume = 0;
		else
			v.volume = 0xFFFF;
		v.bass = 0;
		v.treble = 0;
		v.flags = (((pcr->flags & VIDEO_AUDIO_MUTE)?VIDEO_AUDIO_MUTE:0) |
				VIDEO_AUDIO_MUTABLE);
		strcpy(v.name, "Radio");
		v.mode = VIDEO_SOUND_STEREO;
		v.balance = 32768;
		v.step = 0xFFFF;
		if (copy_to_user(arg,&v, sizeof(v)))
			return -EFAULT;
		return 0;			
	}
	case VIDIOCSAUDIO:
	{
		struct video_audio v;

		if (copy_from_user(&v, arg, sizeof(v)))
			return -EFAULT;
		if (v.audio)
			return -EINVAL;
		/* only mute/unmute */
		if ((v.flags & VIDEO_AUDIO_MUTE) || (v.volume == 0))
			PCR_MUTE(pcr);
		else
			PCR_UNMUTE(pcr);
		pcr_write(pcr);
		return 0;
	}
	default:
		return -ENOIOCTLCMD;
	}
}

static int pcr_open(struct video_device *dev, int flags)
{
	if (users)
		return -EBUSY;
	users++;
	MOD_INC_USE_COUNT;
	return 0;
}

static void pcr_close(struct video_device *dev)
{
	users--;
	MOD_DEC_USE_COUNT;
}

static struct pcr_device pcr_unit;

static struct video_device pcr_radio=
{
	owner:		THIS_MODULE,
	name:		"SF64PCR radio",
	type:		VID_TYPE_TUNER,
	hardware:	VID_HARDWARE_SF64PCR,
	open:		pcr_open,
	close:		pcr_close,
	ioctl:		pcr_ioctl
};


static int __init pcr_init(void)
{
	struct pci_dev *dev = NULL;
	unsigned int port;
        
	dev=pci_find_device(SF64PCR_VENDOR, SF64PCR_DEVICE, dev);
	if (dev == NULL) {
		printk(KERN_ERR "sf64pcr: card not found.\n");
		return -1;
	}

	pci_read_config_dword(dev, PCI_BASE_ADDRESS_0, &port);
	port &= 0xFFFFFFF80;
	pcr_unit.port = port + 0x52;
	
	pcr_unit.freq = SF64PCR_MINFREQ;
	pcr_unit.stereo = TEA5757_STEREO;
	pcr_unit.lock = TEA5757_S030;
	pcr_unit.band = TEA5757_BAND_FM;
	pcr_unit.search = TEA5757_SEARCH_UP;
	pcr_unit.flags = VIDEO_TUNER_LOW | VIDEO_TUNER_STEREO_ON;
	PCR_MUTE(&pcr_unit);
        
	init_MUTEX(&lock);

	pcr_radio.priv = &pcr_unit;
	if (video_register_device(&pcr_radio, VFL_TYPE_RADIO, radio_nr) == -1)
		return -EINVAL;
        
        pcr_write(&pcr_unit);
	return 0;
}

void __exit pcr_cleanup_module(void)
{
	PCR_MUTE(&pcr_unit);
        video_unregister_device(&pcr_radio);
}


MODULE_AUTHOR("Martin \"Marble\" Beranek, marble@matfyz.cz");
MODULE_DESCRIPTION("Driver for the SF64-PCR (alias PCI RadioLINK) radio card.");

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,10)
MODULE_LICENSE("GPL");
#endif

MODULE_PARM(radio_nr, "i");
MODULE_PARM_DESC(radio_nr, "Number of /dev/video[x] device.");

EXPORT_NO_SYMBOLS;

module_init(pcr_init)
module_exit(pcr_cleanup_module)

